"""
Copyright 2022 Andy Qin. All Rights Reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from .pond_class import Pond as Pond
from .pooled_object import PooledObject as PooledObject
from .pooled_object_factory import PooledObjectFactory as PooledObjectFactory



from typing import Any

class PooledObject:
    create_time: float
    last_borrow_time: float
    keeped_object: Any
    def __init__(self, obj: Any) -> None: ...
    def use(self) -> Any: ...
    def update_brrow_time(self) -> PooledObject: ...



"""
Copyright 2022 Andy Qin. All Rights Reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import array
import hashlib
import math
import typing as t
from threading import RLock, Thread

import madoka


class CountMinSketch(object):
    """
    A class for counting hashable items using the Count-min Sketch strategy.
    It fulfills a similar purpose than `itertools.Counter`.
    The Count-min Sketch is a randomized data structure that uses a constant
    amount of memory and has constant insertion and lookup times at the cost
    of an arbitrarily small overestimation of the counts.
    It has two parameters:
     - `m` the size of the hash tables, larger implies smaller overestimation
     - `d` the amount of hash tables, larger implies lower probability of
           overestimation.
    An example usage:
        from countminsketch import CountMinSketch
        sketch = CountMinSketch(1000, 10)  # m=1000, d=10
        sketch.add("oh yeah")
        sketch.add(tuple())
        sketch.add(1, value=123)
        print sketch["oh yeah"]       # prints 1
        print sketch[tuple()]         # prints 1
        print sketch[1]               # prints 123
        print sketch["non-existent"]  # prints 0
    Note that this class can be used to count *any* hashable type, so it's
    possible to "count apples" and then "ask for oranges". Validation is up to
    the user.
    """

    def __init__(self, m: int, d: int):
        """`m` is the size of the hash tables, larger implies smaller
        overestimation. `d` the amount of hash tables, larger implies lower
        probability of overestimation.
        """
        if not m or not d:
            raise ValueError(
                "Table size (m) and amount of hash functions (d)" " must be non-zero"
            )
        self.m: int = m
        self.d: int = d
        self.sketch = madoka.Sketch(width=m, k=d, max_value=8)
        self.lock = False

    def add(self, x: t.Hashable) -> None:
        """
        Count element `x` as if had appeared `value` times.
        Effectively counts `x` as occurring once.
        """
        count = self.sketch[x]
        self.sketch[x] = count + 1
        if count >= 8:
            if not self.lock:
                self._clean()

    def _clean(self) -> None:
        self.sketch.filter(lambda x: math.floor(x / 2))

    def query(self, x: t.Hashable) -> int:
        """
        Return an estimation of the amount of times `x` has ocurred.
        The returned value always overestimates the real value.
        """
        return self.sketch.get(x)

    def __getitem__(self, x: t.Hashable) -> int:
        """
        A convenience method to call `query`.
        """
        return self.sketch.get(x)



"""
Copyright 2022 Andy Qin. All Rights Reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
# https://peps.python.org/pep-0563/
from __future__ import annotations

import time
from typing import Any


class PooledObject:
    create_time: float
    last_borrow_time: float
    keeped_object: Any

    def __init__(self, obj: Any) -> None:
        self.create_time = time.time()
        self.last_borrow_time = time.time()
        self.keeped_object = obj

    def use(self) -> Any:
        return self.keeped_object

    def update_brrow_time(self) -> PooledObject:
        self.last_borrow_time = time.time()
        return self



import abc
from .pooled_object import PooledObject as PooledObject
from _typeshed import Incomplete

class PooledObjectFactory(metaclass=abc.ABCMeta):
    pooled_maxsize: Incomplete
    least_one: Incomplete
    def __init__(self, pooled_maxsize: int = ..., least_one: bool = ...) -> None: ...
    @abc.abstractmethod
    def creatInstantce(self) -> PooledObject: ...
    @abc.abstractmethod
    def destroy(self, pooled_object: PooledObject) -> None: ...
    @abc.abstractmethod
    def reset(self, pooled_object: PooledObject) -> PooledObject: ...
    @abc.abstractmethod
    def validate(self, pooled_object: PooledObject) -> bool: ...
    def factory_name(self) -> str: ...



import time

import pytest

from pond import Pond, PooledObject, PooledObjectFactory


class Dog:
    name: str
    validate_result: bool = True


class PooledDogFactory(PooledObjectFactory):
    def creatInstantce(self) -> PooledObject:
        dog = Dog()
        dog.name = "puppy"
        return PooledObject(dog)

    def destroy(self, pooled_object: PooledObject) -> None:
        del pooled_object

    def reset(self, pooled_object: PooledObject) -> PooledObject:
        pooled_object.keeped_object.name = "puppy"
        return pooled_object

    def validate(self, pooled_object: PooledObject) -> bool:
        return pooled_object.keeped_object.validate_result


pooled_maxsize = 10
pond = Pond(
    borrowed_timeout=2,
    time_between_eviction_runs=-1,
    thread_daemon=True,
    eviction_weight=0.8,
)
factory = PooledDogFactory(pooled_maxsize=10, least_one=False)

pond.register(factory)
T1 = time.perf_counter()
for i in range(100000):
    pooled_object: PooledObject = pond.borrow(factory)
    dog: Dog = pooled_object.use()
    pond.recycle(pooled_object, factory)
T2 = time.perf_counter()
print("程序运行时间:%s毫秒" % ((T2 - T1) * 1000))



import time
from distutils.log import debug

import pytest

from pond import Pond, PooledObject, PooledObjectFactory


class Dog:
    name: str
    validate_result: bool = True


class PooledDogFactory(PooledObjectFactory):
    def creatInstantce(self) -> PooledObject:
        dog = Dog()
        dog.name = "puppy"
        return PooledObject(dog)

    def destroy(self, pooled_object: PooledObject) -> None:
        del pooled_object

    def reset(self, pooled_object: PooledObject) -> PooledObject:
        pooled_object.keeped_object.name = "puppy"
        return pooled_object

    def validate(self, pooled_object: PooledObject) -> bool:
        return pooled_object.keeped_object.validate_result


pooled_maxsize = 10
pond = Pond(
    borrowed_timeout=2,
    time_between_eviction_runs=300,
    thread_daemon=True,
    eviction_weight=0.8,
)
factory = PooledDogFactory(pooled_maxsize=10, least_one=False)


@pytest.mark.run(order=1)
def test_register() -> None:
    assert factory.pooled_maxsize == pooled_maxsize
    pond.register(factory)


@pytest.mark.run(order=1)
def test_register_by_name() -> None:
    pond.register(factory, name="PuppyFactory")
    assert pond.size() == 2


@pytest.mark.run(order=1)
def test_contains() -> None:
    assert pond.contains(name="PuppyFactory")
    assert pond.contains(name="Notfound") is False


@pytest.mark.run(order=2)
def test_pooled_object_size() -> None:
    assert pond.pooled_object_size(factory) == pooled_maxsize


@pytest.mark.run(order=2)
def test_pooled_is_full() -> None:
    assert pond.is_full(factory)


@pytest.mark.run(order=2)
def test_pooled_is_empty() -> None:
    assert pond.is_empty(factory) == False


@pytest.mark.run(order=2)
def test_counter() -> None:
    pond.recycle(pond.borrow(factory), factory)
    assert pond.counter[factory.factory_name()] == 1
    pond.recycle(pond.borrow(factory), factory)
    assert pond.counter[factory.factory_name()] == 2


@pytest.mark.run(order=2)
def test_count_total_objects() -> None:
    assert pond.count_total_objects() == 20


@pytest.mark.run(order=3)
def test_borrow_and_recycle() -> None:
    pooled_object: PooledObject = pond.borrow(factory)
    dog: Dog = pooled_object.use()
    assert dog.name == "puppy"
    assert pond.pooled_object_size(factory) == pooled_maxsize - 1
    assert pooled_object.keeped_object is dog
    pond.recycle(pooled_object, factory)
    assert pond.pooled_object_size(factory) == pooled_maxsize


@pytest.mark.run(order=3)
def test_validate() -> None:
    pond.register(factory, name="test_validate")
    pond.clear(name="test_validate")
    dog = Dog()
    dog.validate_result = False
    pooled_object: PooledObject = PooledObject(dog)
    pond.recycle(pooled_object, name="test_validate")
    assert pond.pooled_object_size(name="test_validate") == 1
    pooled_object = pond.borrow(name="test_validate")
    assert pooled_object.keeped_object.validate_result


@pytest.mark.run(order=3)
def test_borrow_and_recycle_by_name() -> None:
    pooled_object: PooledObject = pond.borrow(name="PuppyFactory")
    dog: Dog = pooled_object.use()
    assert dog.name == "puppy"
    assert pond.pooled_object_size(name="PuppyFactory") == pooled_maxsize - 1
    assert pooled_object.keeped_object is dog
    pond.recycle(pooled_object, name="PuppyFactory")
    assert pond.pooled_object_size(name="PuppyFactory") == pooled_maxsize


@pytest.mark.run(order=3)
def test_recycle_and_reset() -> None:
    pooled_object: PooledObject = pond.borrow(factory)
    dog: Dog = pooled_object.use()
    assert dog.name == "puppy"
    dog.name = "dinosaurs"
    pond.recycle(pooled_object, factory)
    pooled_object = pond.borrow(factory)
    dog = pooled_object.use()
    assert dog.name == "puppy"
    pond.recycle(pooled_object, factory)


@pytest.mark.run(order=3)
def test_borrowed_timeout() -> None:
    pooled_object: PooledObject = pond.borrow(factory)
    dog: Dog = pooled_object.use()
    assert dog.name == "puppy"
    assert pond.pooled_object_size(factory) == pooled_maxsize - 1
    time.sleep(2)
    pond.recycle(pooled_object, factory)
    assert pond.pooled_object_size(factory) == pooled_maxsize - 1


@pytest.mark.run(order=3)
def test_borrow_when_queue_empty() -> None:
    for i in range(pooled_maxsize):
        pond.borrow(factory)
    assert pond.pooled_object_size(factory) == 0
    pooled_object = pond.borrow(factory)
    assert pooled_object
    pond.recycle(pooled_object, factory)
    assert pond.pooled_object_size(factory) > 0


@pytest.mark.run(order=4)
async def test_eviction() -> None:
    pond._Pond__time_between_eviction_runs = 1  # type: ignore
    await pond._Pond__eviction(debug=True)  # type: ignore


@pytest.mark.run(order=998)
def test_clear() -> None:
    assert pond.pooled_object_size(factory) > 0
    pond.clear(factory)
    assert pond.pooled_object_size(factory) == 0


@pytest.mark.run(order=999)
def test_stop() -> None:
    pond.stop()



"""
Copyright 2022 Andy Qin. All Rights Reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import abc

from .pooled_object import PooledObject


class PooledObjectFactory(metaclass=abc.ABCMeta):
    def __init__(self, pooled_maxsize: int = 8, least_one: bool = False) -> None:
        self.pooled_maxsize = pooled_maxsize
        self.least_one = least_one

    @abc.abstractmethod
    def creatInstantce(self) -> PooledObject:
        pass

    @abc.abstractmethod
    def destroy(self, pooled_object: PooledObject) -> None:
        pass

    @abc.abstractmethod
    def reset(self, pooled_object: PooledObject) -> PooledObject:
        pass

    @abc.abstractmethod
    def validate(self, pooled_object: PooledObject) -> bool:
        pass

    def factory_name(self) -> str:
        return self.__class__.__qualname__



"""
Copyright 2022 Andy Qin. All Rights Reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import asyncio
import math
import time
from asyncio import AbstractEventLoop
from collections import deque
from threading import RLock, Thread
from typing import TYPE_CHECKING, Any, Coroutine, Deque, Dict, Final, Optional

from .count_min_sketch import CountMinSketch
from .pooled_object import PooledObject
from .pooled_object_factory import PooledObjectFactory


class Pond(object):
    def __init__(
        self,
        borrowed_timeout: int = 60,
        time_between_eviction_runs: int = 300,
        eviction_weight: float = 0.8,
        thread_daemon: bool = True,
        loop: Optional[AbstractEventLoop] = None,
    ) -> None:
        """Pond is a high performance object-pooling library for Python, it has
            a smaller memory usage and a higher hit rate.For more details,
            see our user's guide or my blog(https://qin.news).

        Args:
            borrowed_timeout (int, optional): The maximum duration of the
                borrowed object. Defaults to 60.
            time_between_eviction_runs (int, optional): The interval for
                automatic recycling. Defaults to 300. If its value is -1,
                the recycling is turned off.
            eviction_weight (float, optional): Automatic recycling weight.
                Defaults to 0.8.
            thread_daemon (bool, optional): A boolean value indicating whether
                the pond's thread is a daemon thread. Defaults to True.
        """
        self.__borrowed_timeout: int = borrowed_timeout
        if loop is not None:
            self.__loop = loop
        else:
            self.__loop = asyncio.new_event_loop()
        self.__time_between_eviction_runs = time_between_eviction_runs
        self.__eviction_weight = eviction_weight
        self.__async_lock = asyncio.Lock()
        self.__sync_lock = RLock()
        self.__class_dict: Final[Dict[str, PooledObjectFactory]] = dict()

        if TYPE_CHECKING:
            self.__pooled_object_tree: Final[Dict[str, Deque[PooledObject]]] = dict()
        else:
            self.__pooled_object_tree: Final[Dict[str, Deque]] = dict()
        self.counter: CountMinSketch = CountMinSketch(28, 3)

        def loop_runner(
            loop: AbstractEventLoop, function: Coroutine[Any, Any, None]
        ) -> None:
            asyncio.set_event_loop(loop)
            loop.create_task(function)
            loop.run_forever()

        if self.__time_between_eviction_runs > -1:
            self.__thread = Thread(
                target=loop_runner, args=(self.__loop, self.__eviction())
            )
            self.__thread.daemon = thread_daemon
            self.__thread.start()

    def register(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> None:
        """Registering the factory object with Pond will use the class name of
            the factory as the key for the PooledObjectTree by default.After
            successful registration, Pond will automatically start creating
            objects based on the pooled_maxsize set in factory until the pool
            is filled.

        Args:
            factory (Optional[PooledObjectFactory], optional): The factory
                object you want to register. Defaults to None.
            name (Optional[str], optional): The factory name you want to register.
                Defaults to None.

        Raises:
            ValueError: The factoryClass existed in the PooledObjectTree!
            ValueError: The instance must not be null!
        """
        if name is None:
            assert factory is not None
            name = factory.factory_name()
        assert factory is not None
        if self.__pooled_object_tree.__contains__(name):
            raise ValueError("The factoryClass existed in the PooledObjectTree!")
        self.__class_dict[name] = factory
        self.__pooled_object_tree[name] = deque(maxlen=factory.pooled_maxsize)
        for i in range(factory.pooled_maxsize):
            instance = self.__class_dict[name].creatInstantce()
            if instance is None:
                raise ValueError("The instance must not be null!")
            self.__pooled_object_tree[name].appendleft(instance)

    def borrow(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> PooledObject:
        """You can use factory object or factory name to borrow and return
            objects from the object pool.

        Args:
            factory (Optional[PooledObjectFactory], optional): The factory
                object you want to register. Defaults to None.
            name (Optional[str], optional): The factory name you want to register.
                Defaults to None.

        Returns:
            PooledObject: The pooled object you want to borrow.
        """
        with self.__sync_lock:
            if not name:
                assert factory is not None
                name = factory.factory_name()
            if self.is_empty(name=name):
                if self.__time_between_eviction_runs > -1:
                    self.counter.add(name)
                return self.__class_dict[name].creatInstantce()
            pooled_object = self.__pooled_object_tree[name].pop()
            while not self.__class_dict[name].validate(pooled_object):
                self.__clear_one_object(pooled_object, name=name)
                if self.is_empty(name=name):
                    pooled_object = self.__class_dict[name].creatInstantce()
                else:
                    pooled_object = self.__pooled_object_tree[name].pop()
            if self.__time_between_eviction_runs > -1:
                self.counter.add(name)
            return pooled_object.update_brrow_time()

    def recycle(
        self,
        pooled_object: PooledObject,
        factory: Optional[PooledObjectFactory] = None,
        name: Optional[str] = None,
    ) -> None:
        """You can use factory object or factory name to borrow and return
            objects from the object pool.

        Args:
            factory (Optional[PooledObjectFactory], optional): The factory
                object you want to register. Defaults to None.
            name (Optional[str], optional): The factory name you want to register.
                Defaults to None.
        """
        with self.__sync_lock:
            if not name:
                assert factory is not None
                name = factory.factory_name()
            if not isinstance(pooled_object, PooledObject):
                raise ValueError("Only PooledObject can be recycled!")
            if self.is_full(name=name):
                self.__clear_one_object(pooled_object, name=name)
                return
            if (time.time() - pooled_object.last_borrow_time) > self.__borrowed_timeout:
                self.__clear_one_object(pooled_object, name=name)
                return

            self.__pooled_object_tree[name].appendleft(
                self.__class_dict[name].reset(pooled_object)
            )

    def clear(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> None:
        """You can use factory object or factory name to clear the object pool.

        Args:
            factory (Optional[PooledObjectFactory], optional): The factory
                object you want to register. Defaults to None.
            name (Optional[str], optional): The factory name you want to register.
                Defaults to None.
        """
        if not name:
            assert factory is not None
            name = factory.factory_name()
        while not self.is_empty(name=name):
            pooled_object = self.__pooled_object_tree[name].pop()
            self.__clear_one_object(pooled_object, name=name)

    def __clear_one_object(
        self,
        pooled_object: PooledObject,
        factory: Optional[PooledObjectFactory] = None,
        name: Optional[str] = None,
    ) -> None:
        if not name:
            assert factory is not None
            name = factory.factory_name()
        self.__class_dict[name].destroy(pooled_object)
        del pooled_object

    def size(self) -> int:
        """Query how many object pools there are in pooled_object_tree."""
        return len(self.__pooled_object_tree)

    def count_total_objects(self) -> int:
        """Query how many objects there are in pooled_object_tree."""
        total_number = 0
        for k, v in self.__pooled_object_tree.items():
            total_number = total_number + len(v)
        return total_number

    def contains(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> bool:
        """Return true if the specified factory has been registered in the pond.

        Args:
            factory (Optional[PooledObjectFactory], optional): The specified factory object. Defaults to None.
            name (Optional[str], optional): The specified factory name. Defaults to None.

        Returns:
            bool: Return true if the specified factory has been registered in the pond.
        """
        if name is None:
            assert factory is not None
            name = factory.factory_name()
        return self.__pooled_object_tree.__contains__(name)

    def pooled_object_size(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> int:
        """Query how many objects there are in the specified object pool.

        Args:
            factory (Optional[PooledObjectFactory], optional): The specified factory object. Defaults to None.
            name (Optional[str], optional): The specified factory name. Defaults to None.

        Returns:
            int: Size of the specified object pool.
        """
        if not name:
            assert factory is not None
            name = factory.factory_name()
        return len(self.__pooled_object_tree[name])

    def is_full(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> bool:
        """Check to see if the supplied object pool is filled.

        Args:
            factory (Optional[PooledObjectFactory], optional): The specified factory object. Defaults to None.
            name (Optional[str], optional): The specified factory name. Defaults to None.

        Returns:
            bool: Whether the object pool is full.
        """
        if not name:
            assert factory is not None
            name = factory.factory_name()
        return len(self.__pooled_object_tree[name]) >= (
            self.__pooled_object_tree[name].maxlen or 0
        )

    def is_empty(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> bool:
        """Check to see if the supplied object pool is emptied.

        Args:
            factory (Optional[PooledObjectFactory], optional): The specified factory object. Defaults to None.
            name (Optional[str], optional): The specified factory name. Defaults to None.

        Returns:
            bool: Whether the object pool is empty.
        """
        if not name:
            assert factory is not None
            name = factory.factory_name()
        return len(self.__pooled_object_tree[name]) == 0

    def __reset_counter(self) -> None:
        n = self.size()
        if n == 0:
            epsilon = 0.1
        else:
            epsilon = 10 / n
        if epsilon >= 1:
            epsilon = 0.1
        m = math.ceil(math.e / epsilon)
        d = math.ceil(math.log(1 / 0.1))
        if not m == self.counter.m or not d == self.counter.d:
            self.counter = CountMinSketch(m, d)

    def stop(self) -> None:
        """Stop the pone and all objects in the pooled object tree will be destroyed."""
        self.__loop.stop()
        self.__time_between_eviction_runs = -1
        for key in self.__pooled_object_tree.keys():
            self.clear(name=key)

    async def __eviction(self, debug: bool = False) -> None:
        first_run = True
        while self.__time_between_eviction_runs > 0:
            if first_run:
                first_run = False
                await asyncio.sleep(self.__time_between_eviction_runs)
                continue
            pooled_object_borrow_count: Dict[str, int] = {}
            max_count = 8
            for key in self.__pooled_object_tree.keys():
                pooled_object_borrow_count[key] = self.counter[key]
            boundary = int(max_count * self.__eviction_weight)
            for key, value in pooled_object_borrow_count.items():
                size = len(self.__pooled_object_tree[key])
                if value < boundary and size > 0:
                    if size > 1:
                        for i in range(int(size / 2)):
                            self.__clear_one_object(
                                self.__pooled_object_tree[key].pop(),
                                name=key,
                            )
                    else:
                        if not self.__class_dict[key].least_one:
                            self.clear(name=key)
            self.__reset_counter()
            if debug:
                self.__time_between_eviction_runs = -1
            await asyncio.sleep(self.__time_between_eviction_runs)

    async def async_register(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> None:
        async with self.__async_lock:
            self.register(factory=factory, name=name)

    async def async_borrow(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> PooledObject:
        async with self.__async_lock:
            return self.borrow(factory=factory, name=name)

    async def async_recycle(
        self,
        pooled_object: PooledObject,
        factory: Optional[PooledObjectFactory] = None,
        name: Optional[str] = None,
    ) -> None:
        async with self.__async_lock:
            self.recycle(pooled_object=pooled_object, factory=factory, name=name)

    async def async_clear(
        self, factory: Optional[PooledObjectFactory] = None, name: Optional[str] = None
    ) -> None:
        async with self.__async_lock:
            self.clear(factory=factory, name=name)



import asyncio
import time

import pytest

from pond import Pond, PooledObject, PooledObjectFactory


class Dog:
    name: str
    validate_result: bool = True


class PooledDogFactory(PooledObjectFactory):
    def creatInstantce(self) -> PooledObject:
        dog = Dog()
        dog.name = "puppy"
        return PooledObject(dog)

    def destroy(self, pooled_object: PooledObject) -> None:
        del pooled_object

    def reset(self, pooled_object: PooledObject) -> PooledObject:
        pooled_object.keeped_object.name = "puppy"
        return pooled_object

    def validate(self, pooled_object: PooledObject) -> bool:
        return pooled_object.keeped_object.validate_result


pooled_maxsize = 10
pond = Pond(
    borrowed_timeout=2,
    time_between_eviction_runs=-1,
    thread_daemon=True,
    eviction_weight=0.8,
)
factory = PooledDogFactory(pooled_maxsize=10, least_one=False)


@pytest.mark.run(order=1)
async def test_async_borrow_and_recycle() -> None:
    await pond.async_register(factory)
    pooled_object: PooledObject = await pond.async_borrow(factory)
    dog: Dog = pooled_object.use()
    assert dog.name == "puppy"
    assert pond.pooled_object_size(factory) == pooled_maxsize - 1
    assert pooled_object.keeped_object is dog
    await pond.async_recycle(pooled_object, factory)
    assert pond.pooled_object_size(factory) == pooled_maxsize



from .count_min_sketch import CountMinSketch as CountMinSketch
from .pooled_object import PooledObject as PooledObject
from .pooled_object_factory import PooledObjectFactory as PooledObjectFactory
from _typeshed import Incomplete
from asyncio import AbstractEventLoop
from typing import Optional

class Pond:
    counter: Incomplete
    def __init__(self, borrowed_timeout: int = ..., time_between_eviction_runs: int = ..., eviction_weight: float = ..., thread_daemon: bool = ..., loop: Optional[AbstractEventLoop] = ...) -> None: ...
    def register(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> None: ...
    def borrow(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> PooledObject: ...
    def recycle(self, pooled_object: PooledObject, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> None: ...
    def clear(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> None: ...
    def size(self) -> int: ...
    def count_total_objects(self) -> int: ...
    def contains(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> bool: ...
    def pooled_object_size(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> int: ...
    def is_full(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> bool: ...
    def is_empty(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> bool: ...
    def stop(self) -> None: ...
    async def async_register(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> None: ...
    async def async_borrow(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> PooledObject: ...
    async def async_recycle(self, pooled_object: PooledObject, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> None: ...
    async def async_clear(self, factory: Optional[PooledObjectFactory] = ..., name: Optional[str] = ...) -> None: ...



import typing as t
from _typeshed import Incomplete
from threading import RLock as RLock, Thread as Thread

class CountMinSketch:
    m: Incomplete
    d: Incomplete
    sketch: Incomplete
    lock: bool
    def __init__(self, m: int, d: int) -> None: ...
    def add(self, x: t.Hashable) -> None: ...
    def query(self, x: t.Hashable) -> int: ...
    def __getitem__(self, x: t.Hashable) -> int: ...
